import java.util.*;
import java.awt.*;

/**
 * Clase que representa una bola en el simulador.
 * Describe su simulación a traves del tiempo y al aplicarle fuerzas externas.
 *
 * @version 0.4
*/
@SuppressWarnings("serial")
public class Ball extends PhysicsElement implements Simulateable  {
	private static int id=0;  // Ball identification number
	private final double mass;
	private final double radius;
	private Vector2D pos_t;				// current position at time t
	private Vector2D pos_tPlusDelta;	// next position at t + delta
	private Vector2D speed_t;			// speed at time t
	private Vector2D speed_tPlusDelta;	// speed at t + delta
	private Vector2D a_t;				// acceleration at time t
	private Vector2D a_tMinusDelta; 
	private ArrayList<Spring> springs;
	private BallView view;

	/**
	* Constructor privado.
	* 
	*/
	private Ball(){	// nobody can create a ball without state
		super(id++);
		mass = 0;
		radius = 0;
	}
	/**
	* Constructor público.
	* 
	* Inicializa el objeto y crea vista para desplegarla en pantalla.
	* @see BallView
	* 
	* @param mass 		Masa de la bola [Kg]
	* @param r 			Radio de la bola [m]
	* @param position 	Posición de la bola en cordenadas cartesianas (x [m], y [m]) 
	* @param speed 		Velocidad inicial de la bola visto como vector (Vx [m/s], Vy [m/s])
	*/
	public Ball(double mass, double r, Vector2D position, Vector2D speed){
		super(id++);
		this.mass = mass;
		radius = r;
		pos_t = position;
		speed_t = speed;
		a_t=a_tMinusDelta= new Vector2D();
		springs = new ArrayList<Spring>();
		view = new BallView(this);
	}
	/**
	* Engancha un resorte a la bola. La bola puede tener múltiples resortes.
	* @param s 		Resorte que se une a la bola
	*/
	public void attach(Spring s){
		springs.add(s);
	}
	/**
	* Desengancha un resorte a la bola
	* @param s 		Resorte que se desengancha de la bola
	*/
	public void detach(Spring s){
		springs.remove(s);
	}
	/**
	* Retorna el numero de identificación de la bola
	* @return id 	id de la bola
	*/
	public double getID() {
		return id;
	}
	/**
	* Retorna la masa de la bola
	* @return mass 		Masa de la bola en [Kg]
	*/
	public double getMass() {
		return mass;
	}
	/**
	* Retorna el radio de la bola
	* @return radius 	Radio de la bola en [m]
	*/
	public double getRadius() {
		return radius;
	}
	/**
	* Retorna la posición actual de la bola en el plano cartesiano
	* @return pos_t 	Posición actual de la bola en (X, Y) en [m] 
	*/
	public Vector2D getPosition() {
		return pos_t;
	}
	/**
	* Retorna la velocidad actual de la bola en forma vectorial
	* @return speed_t 	 Velocidad actual de la bola en (Vx, Vy) en [m/s]
	*/
	public Vector2D getSpeed() {
		return speed_t;
	}
	/**
	* Método para obtener la suma de todas las fuerzas ejercidas sobre la bola.
	* @return force 	Fuerza neta en la bola [N]
	*/
	private Vector2D getNetForce(){
		Vector2D force = new Vector2D();
		for (Spring s:springs)
			force=force.plus(s.getForce(this));
		return force;
	}


	//******************************************
	//***** MÉTODOS OBLIGADOS POR HERENCIA *****
	//******************************************
	/**
	* Calcula si dos bolas estan colisionando
	* @param b 			Bola con la que se compara
	* @return True 		Si existe colision entre la bola que invoca el método y la bola 'b'. False 	Si no existe colision entre las bolas.
	*/
	public PhysicsElement collide(Ball b){
		if (this == b)
			return null;
		Vector2D vectorBetweenCenters=getPosition().minus(b.getPosition());
		boolean areTouching = vectorBetweenCenters.module() < radius+b.getRadius();
		Vector2D diffSpeed = getSpeed().minus(b.getSpeed());
		boolean areApproaching = diffSpeed.dot(vectorBetweenCenters) < 0;
		return (areTouching && areApproaching?this: null);
	}
	/**
	* Redibuja la bola.
	* @param g 	Objeto Graphics2D para manejar la vista	
	* @see MyWorldView
	*/	
	public void paintView(Graphics2D g){
		view.draw(g);
	}
	/**
	* Obtiene la descripción de la bola. 
	* @return String de la forma "Ball_"+getId()+":x"
	*/
	public String getDescription() {
		return "Ball_" + getId()+Vector2D.getDescription();
	}
	/**
	* Obtiene el estado actual (Posición Actual) de la bola en la simulación. 
	* @return String de la forma getPosition()+" "
	*/
	public String getState() {
		return getPosition()+" ";
	}
	/**
	* Verifica si un punto en el plano cartesiano se encuentra en la sección de selección de la bola
	* @param x 		Punto en el eje x
	* @param y 		Punto en el eje y
	* @return True 	Si el punto (x, y) se encuentra en la sección de selección. False 	Si el punto no esta contenido en la sección
	*/
	public boolean contains (double x, double y)  {
		return (getPosition().minus(new Vector2D( x, y ))).module() < 0.5*getRadius();
	}
	/**
	* Mueve a la bola en el plano cartesiano según las distancias recibidas
	* @param dx 	Distancia que se movio la bola en el eje X en [m]
	* @param dy 	Distancia que se movio la bola en el eje Y en [m]
	*/
	public void dragDelta(double dx, double dy)  {
		pos_t = pos_t.plus( new Vector2D(dx, dy) );
	}
	/**
	* Marca a la bola como seleccionada para manipulación. La pinta de color rojo.
	* @see MouseListener
	*/
	public void setSelected()  {
		view.color = Color.RED;
	}
	/**
	* Libera a la bola como seleccionada para manipulación. Devuelve la bola a su color por defecto (verde).
	* @see MouseListener
	*/
	public void setReleased()  {
		view.color = Color.GREEN;
	}

	//******************************************
	//***** MÉTODOS OBLIGADOS POR INTERFAZ *****
	//******************************************

	/**
	* Calcula el estado siguiente de la bola respecto al mundo
	* @param delta_t 	Diferencia de tiempo entre el estado de simulación actual y el futuro en [s]
	* @param w 			Referencia al mundo que contiene la bola
	* @see Simulateable
	* @see MyWorld
	*/
	public void computeNextState(double delta_t, MyWorld w) {
		Vector2D speed=speed_t;
		PhysicsElement pe;
		if ((pe=w.getCollidingElement(this))!= null){ /* elastic collision */
			if (pe instanceof Ball) {
				Ball b =(Ball) pe;
				Vector2D dpos = getPosition().minus(b.getPosition()).unitary();
				Vector2D dv = getSpeed().minus(b.getSpeed());
				speed=speed_t.minus( dpos.times( dv.dot(dpos)*2*b.getMass()/(mass+b.getMass()) ) );
			}
			else { 
				if( pe instanceof Wall)
					if (((Wall)pe).getOrientation()==Wall.SIDEWAYS)
						speed = new Vector2D(speed_t.getX(),-speed_t.getY());
					else
						speed = new Vector2D(-speed_t.getX(),speed_t.getY());
			}
		} 
		
		a_t = getNetForce().times(1/mass);
		speed_tPlusDelta = speed.plus(a_t.times(3).minus(a_tMinusDelta).times(delta_t/2));
		pos_tPlusDelta = pos_t.plus(speed.times(delta_t)).plus(a_t.times(4).minus(a_tMinusDelta).times(delta_t*delta_t/6));
	}
	/**
	* Actualiza el estado de la bola según los cálculos efectuados en computeNextState
	* @see Simulateable
	*/
	public void updateState(){
		pos_t = pos_tPlusDelta;
		speed_t = speed_tPlusDelta;
		a_tMinusDelta = a_t;
	}
}
